/* positional character frequencies. init to all .25*/

_EFV_MATRIX0_ = {4,4};
_EFV_MATRIX1_ = {4,4};
_EFV_MATRIX2_ = {4,4};

for (h=0; h<4; h=h+1)
{
	for (v=0; v<4; v=v+1)
	{
		_EFV_MATRIX0_ [h][v] = .25;
		_EFV_MATRIX1_ [h][v] = .25;
		_EFV_MATRIX2_ [h][v] = .25;
	}
}

#include "CodonToolsMain.def";
