global beta = .25;
beta:>0.01;beta:<100;
categDef2 = " ";

global Inv_prop;
Inv_prop :< 1-(1e-10);
Inv_prop = 1/resp2;

if (randomizeInitValues)
{
	global beta = Random(0.1,10);
	global Inv_prop = Random (0.05,0.95);
}

catFreqMatrix = {resp2,1};
catFreqMatrix [0] := Inv_prop;

for (h=1; h<resp2;h=h+1)
{
	catFreqMatrix [h] := (1-Inv_prop)/(resp2__-1);
}


if (USE_CONST_ON_PART)
{
category d = (resp2, catFreqMatrix, MEAN, 
				(1-Inv_prop)*GammaDist(_x_,beta,beta)*(_x_>0), 
				(1-Inv_prop)*CGammaDist(_x_,beta,beta)*(_x_>0)+Inv_prop, 
				0, 
		  	    1e25,
		  	    (1-Inv_prop)*CGammaDist(_x_,beta+1,beta)*(_x_>0),
		  	    CONSTANT_ON_PARTITION
		  	 );
}
else
{
category d = (resp2, catFreqMatrix, MEAN, 
				(1-Inv_prop)*GammaDist(_x_,beta,beta)*(_x_>0), 
				(1-Inv_prop)*CGammaDist(_x_,beta,beta)*(_x_>0)+Inv_prop, 
				0, 
		  	    1e25,
		  	    (1-Inv_prop)*CGammaDist(_x_,beta+1,beta)*(_x_>0)
		  	 );

}
		  	 

